@@ -15,6 +15,7 @@ package Net::SIP::Simple::RTP;
 use Net::SIP::Util qw(invoke_callback);
 use Socket;
 use Net::SIP::Debug;
+use Net::SIP::DTMF;
 
 
 # on MSWin32 non-blocking sockets are not supported from IO::Socket
@@ -36,7 +37,8 @@ sub media_recv_echo {
 
 		my $lsocks = $args->{media_lsocks};
 		my $ssocks = $args->{media_ssocks} || $lsocks;
-		my $raddr = $args->{media_raddr};
+		my $raddr  = $args->{media_raddr};
+		my $mdtmf  = $args->{media_dtmfxtract};
 		my $didit = 0;
 		for( my $i=0;1;$i++ ) {
 			my $sock = $lsocks->[$i] || last;
@@ -50,13 +52,22 @@ sub media_recv_echo {
 			my @delay_buffer;
 			my $echo_back = sub {
 				my ($s_sock,$remote,$delay_buffer,$delay,$writeto,$targs,$didit,$sock) = @_;
-				for my $dummy (1) {
-					my $buf = _receive_rtp( $sock,$writeto,$targs,$didit );
-					defined($buf) or last;
+				{
+					my ($buf,$mpt,$seq,$tstamp,$ssrc,$csrc) = 
+						_receive_rtp( $sock,$writeto,$targs,$didit )
+						or last;
 					#DEBUG( "$didit=$$didit" );
 					$$didit = 1;
-					next if $delay<0;
-					next if ! $remote; # call on hold ?
+
+					my @pkt = _generate_dtmf($targs,$seq,$tstamp,0x1234);
+					if (@pkt && $pkt[0] ne '') {
+						DEBUG( 100,"send DTMF to RTP");
+						send( $s_sock,$_,0,$remote ) for(@pkt);
+						return; # send DTMF *instead* of echo data
+					}
+
+					last if $delay<0;
+					last if ! $remote; # call on hold ?
 					push @$delay_buffer, $buf;
 					while ( @$delay_buffer > $delay ) {
 						send( $s_sock,shift(@$delay_buffer),0,$remote );
@@ -66,7 +77,11 @@ sub media_recv_echo {
 			};
 
 			$call->{loop}->addFD( $sock,
-				[ $echo_back,$s_sock,$addr,\@delay_buffer,$delay || 0,$writeto,{},\$didit ],
+				[ $echo_back,$s_sock,$addr,\@delay_buffer,$delay || 0,$writeto,{
+					dtmf_gen => $args->{dtmf_events},
+					dtmf_xtract => $mdtmf && $mdtmf->[$i] && $args->{cb_dtmf} 
+						&& [ $mdtmf->[$i], $args->{cb_dtmf} ],
+				},\$didit ],
 				'rtp_echo_back' );
 			my $reset_to_blocking = CAN_NONBLOCKING && $s_sock->blocking(0);
 			push @{ $call->{ rtp_cleanup }}, [ sub {
@@ -124,7 +139,8 @@ sub media_send_recv {
 
 		my $lsocks = $args->{media_lsocks};
 		my $ssocks = $args->{media_ssocks} || $lsocks;
-		my $raddr = $args->{media_raddr};
+		my $raddr  = $args->{media_raddr};
+		my $mdtmf  = $args->{media_dtmfxtract};
 		my $didit = 0;
 		for( my $i=0;1;$i++ ) {
 			my $sock = $lsocks->[$i] || last;
@@ -157,7 +173,10 @@ sub media_send_recv {
 						repeat => $repeat || 1,
 						cb_done => [ sub { invoke_callback(@_) }, $cb_done, $call ],
 						rtp_param => $args->{rtp_param},
-					}],
+						dtmf_gen => $args->{dtmf_events},
+						dtmf_xtract => $mdtmf && $mdtmf->[$i] && $args->{cb_dtmf} 
+							&& [ $mdtmf->[$i], $args->{cb_dtmf} ],
+						}],
 					$args->{rtp_param}[2], # repeat timer
 					'rtpsend',
 				);
@@ -230,7 +249,7 @@ sub _receive_rtp {
 	}
 	# skip csrc headers
 	my $cc = $vpxcc & 0x0f;
-	substr( $buf,0,4*$cc,'' ) if $cc;
+	my $csrc = $cc && substr( $buf,0,4*$cc,'' );
 
 	# skip extension header
 	my $xh = $vpxcc & 0x10 ? (unpack( 'nn', substr( $buf,0,4,'' )))[1] : 0;
@@ -260,8 +279,15 @@ sub _receive_rtp {
 		}
 		syswrite($fd,$payload);
 	}
+	
+	if ( my $xt = $targs->{dtmf_xtract} ) {
+		my ($sub,$cb) = @$xt;
+		if ( my ($event,$duration) = $sub->($packet)) {
+			$cb->($event,$duration);
+		}
+	}
 
-	return $packet;
+	return wantarray ? ( $packet,$mpt,$seq,$tstamp,$ssrc,$csrc ): $packet;
 }
 
 ###########################################################################
@@ -283,16 +309,33 @@ sub _receive_rtp {
 sub _send_rtp {
 	my ($sock,$loop,$addr,$readfrom,$targs,$timer) = @_;
 
-	my ($buf);
+	$targs->{wseq}++;
+	my $seq = $targs->{wseq};
+
+	# 32 bit timestamp based on seq and packet size
+	my $timestamp = ( $targs->{rtp_param}[1] * $seq ) % 2**32;
+
+	my @pkt = _generate_dtmf($targs,$seq,$timestamp,0x1234);
+	if (@pkt && $pkt[0] ne '') {
+		DEBUG( 100,"send DTMF to RTP");
+		send( $sock,$_,0,$addr ) for(@pkt);
+		return; 
+	}
+
+	my $buf;
+	my $rtp_event;
+	my $payload_type;
+
 	if ( ref($readfrom) ) {
 		# payload by callback
-		$buf = invoke_callback( $readfrom );
+		$buf = invoke_callback($readfrom,$seq);
 		if ( !$buf ) {
 			DEBUG( 50, "no more data from callback" );
 			$timer && $timer->cancel;
 			invoke_callback( $targs->{cb_done} );
 			return;
 		}
+		($buf,$payload_type,$rtp_event,$timestamp) = @$buf if ref($buf);
 	} else {
 		# read from file
 		for(my $tries = 0; $tries<2;$tries++ ) {
@@ -319,15 +362,8 @@ sub _send_rtp {
 			$targs->{repeat}--;
 		}
 	}
-	$buf || die $!;
-
-	# add RTP header
-	$targs->{wseq}++;
-	my $seq = $targs->{wseq};
-	# 32 bit timestamp based on seq and packet size
-	# FIXME: it assumes here that packet size == number of samples which is true only for 8bit
-	my $timestamp = ( $targs->{rtp_param}[1] * $seq ) % 2**32;
 
+	die $! if ! defined $buf or $buf eq '';
 	if (0) {
 		my ($fp,$fa) = unpack_sockaddr_in( getsockname($sock) );
 		$fa = inet_ntoa($fa);
@@ -336,9 +372,14 @@ sub _send_rtp {
 		DEBUG( 50, "$fa:$fp -> $ta:$tp seq=$seq ts=%x",$timestamp );
 	}
 
+	# add RTP header
+	$rtp_event = 0 if ! defined $rtp_event;
+	$payload_type = $targs->{rtp_param}[0]||0   # 0 == PMCU 8000
+		if ! defined $payload_type; 
+
 	my $header = pack('CCnNN',
 		0b10000000, # Version 2
-		$targs->{rtp_param}[0], # 0 == PMCU 8000
+		$payload_type | ( $rtp_event << 7 ) ,
 		$seq, # sequence
 		$timestamp,
 		0x1234,    # source ID
@@ -347,4 +388,37 @@ sub _send_rtp {
 	send( $sock,$header.$buf,0,$addr ) || die $!;
 }
 
+###########################################################################
+# Helper to send DTMF
+# Args: ($targs,$seq,$timestamp,$srcid)
+#  $targs: hash which is shared with _send_rtp and other callbacks, contains
+#    dtmf array with events 
+#  $seq,$timestamp,$srcid: parameter for RTP packet
+# Returns: @pkt
+#  ():            no DTMF events to handle
+#  $pkt[0] eq '': DTMF in process, but no data
+#  @pkt:          RTP packets to send
+###########################################################################
+sub _generate_dtmf {
+	my ($targs,$seq,$timestamp,$srcid) = @_;
+	my $dtmfs = $targs->{dtmf_gen};
+	$dtmfs and @$dtmfs or return;
+	DEBUG(100,"got %d dtmfs",0+@{$dtmfs||[]});
+
+	while ( @$dtmfs ) {
+		my $dtmf = $dtmfs->[0];
+		if ( my $duration = $dtmf->{duration} ) {
+			my $cb = $dtmf->{sub} 
+				||= dtmf_generator($dtmf->{event},$duration,%$dtmf);
+			my @pkt = $cb->($seq,$timestamp,$srcid);
+			return @pkt if @pkt;
+		}
+		shift(@$dtmfs);
+		if ( my $cb = $dtmf->{cb_final} ) {
+			invoke_callback($cb,'OK');
+		}
+	}
+	return;
+}
+
 1;
